@@ -760,6 +760,8 @@ class Sequence2Sequence(object):
         perplexity scores, and input-output alignments.
         '''
         assert not fast or greedy, "cannot decode in fast mode with beam search enabled"
+        if not lines:
+            return [], [], [], []
         
         # vectorize:
         encoder_input_data, _, _, _ = self.vectorize_lines(lines, lines, conf)
@@ -790,6 +792,7 @@ class Sequence2Sequence(object):
                         probs = [1.0] * len(line)
                         score = 0
                         alignment = np.eye(len(line)).tolist()
+                line = line.replace(GAP, '') # remove if rejected (i.e. not corrected despite underspecification)
                 output_lines.append(line)
                 output_probs.append(probs)
                 output_scores.append(score)
