@@ -1,11 +1,14 @@
 from __future__ import absolute_import
 
+import os
 import math
+import json
 
 from ocrd import Processor
 from ocrd_utils import (
     getLogger,
     assert_file_grp_cardinality,
+    make_file_id,
     MIMETYPE_PAGE
 )
 from ocrd_modelfactory import page_from_file
@@ -28,14 +31,14 @@ class EvaluateLines(Processor):
         Find files in all input file groups of the workspace for the same
         pageIds. The first file group serves as reference annotation (ground truth).
         
-        Open and deserialise PAGE input files, then iterative over the element
+        Open and deserialise PAGE input files, then iterate over the element
         hierarchy down to the TextLine level, looking at each first TextEquiv.
         Align character sequences in all pairs of lines for the same TextLine IDs,
         and calculate the distances using the error metric `metric`. Accumulate
         distances and sequence lengths per file group globally and per file,
         and show each fraction as a CER rate in the log.
         """
-        #assert_file_grp_cardinality(self.output_file_grp, 0)
+        assert_file_grp_cardinality(self.output_file_grp, 1)
 
         LOG = getLogger('processor.EvaluateLines')
 
@@ -46,7 +49,7 @@ class EvaluateLines(Processor):
         ifgs = self.input_file_grp.split(",") # input file groups
         if len(ifgs) < 2:
             raise Exception("need multiple input file groups to compare")
-        ifts = self.zip_input_files(ifgs) # input file tuples
+        ifts = self.zip_input_files(mimetype=MIMETYPE_PAGE) # input file tuples
 
         # get separate aligners (1 more than needed), because
         # they are stateful (confusion counts):
@@ -68,18 +71,22 @@ class EvaluateLines(Processor):
                     continue
                 LOG.info("INPUT FILE for %s: %s", ifgs[i], input_file.ID)
                 pcgts = page_from_file(self.workspace.download_file(input_file))
-                file_lines[i] = _page_get_lines(pcgts)
+                file_lines[i] = page_get_lines(pcgts)
             # compare lines with GT:
+            report = dict()
             for line_id in file_lines[0].keys():
                 for i, input_file in enumerate(ift):
                     if not i:
                         continue
-                    elif not input_file:
+                    pair = ifgs[0] + ',' + ifgs[i]
+                    lines = report.setdefault(pair, dict()).setdefault('lines', list())
+                    if not input_file:
                         # file/page was not found in this group
                         continue
                     elif line_id not in file_lines[i]:
                         LOG.error('line "%s" in file "%s" is missing from input %d',
                                   line_id, input_file.ID, i)
+                        lines.append({line_id: 'missing'})
                         continue
                     gt_line = file_lines[0][line_id]
                     ocr_line = file_lines[i][line_id]
@@ -88,15 +95,19 @@ class EvaluateLines(Processor):
                     if 0.2 * (gt_len + ocr_len) < math.fabs(gt_len - ocr_len) > 5:
                         LOG.warning('line "%s" in file "%s" deviates significantly in length (%d vs %d)',
                                     line_id, input_file.ID, gt_len, ocr_len)
-                    # align and accumulate edit counts for lines:
-                    file_edits[i].add(
+                    if metric == 'Levenshtein-fast':
                         # not exact (but fast): codepoints
-                        self.aligners[i].get_levenshtein_distance(ocr_line, gt_line)
-                        if metric == 'Levenshtein' else
+                        dist = self.aligners[i].get_levenshtein_distance(ocr_line, gt_line)
+                    else:
                         # exact (but slow): grapheme clusters
-                        self.aligners[i].get_adjusted_distance(ocr_line, gt_line,
-                                                               # NFC / NFKC / historic_latin
-                                                               normalization=metric))
+                        dist = self.aligners[i].get_adjusted_distance(ocr_line, gt_line,
+                                                                      # Levenshtein / NFC / NFKC / historic_latin
+                                                                      normalization=metric)
+                    # align and accumulate edit counts for lines:
+                    file_edits[i].add(dist)
+                    # todo: maybe it could be useful to retrieve and store the alignments, too
+                    lines.append({line_id: {'length': gt_len, 'distance': dist}})
+            
             # report results for file
             for i, input_file in enumerate(ift):
                 if not i:
@@ -109,9 +120,26 @@ class EvaluateLines(Processor):
                          file_edits[i].mean,
                          math.sqrt(file_edits[i].varia),
                          input_file.pageId, ifgs[0], ifgs[i])
+                pair = ifgs[0] + ',' + ifgs[i]
+                report[pair]['length'] = file_edits[i].length
+                report[pair]['distance-mean'] = file_edits[i].mean
+                report[pair]['distance-varia'] = file_edits[i].varia
                 # accumulate edit counts for files
                 edits[i].merge(file_edits[i])
+            
+            # write back result to page report
+            file_id = make_file_id(ift[0], self.output_file_grp)
+            file_path = os.path.join(self.output_file_grp, file_id + '.json')
+            self.workspace.add_file(
+                ID=file_id,
+                file_grp=self.output_file_grp,
+                pageId=input_file.pageId,
+                local_filename=file_path,
+                mimetype='application/json',
+                content=json.dumps(report, indent=2))
+            
         # report overall results
+        report = dict()
         for i in range(1, len(ifgs)):
             if not edits[i].length:
                 LOG.warning('%s had no textlines whatsoever', ifgs[i])
@@ -121,35 +149,31 @@ class EvaluateLines(Processor):
                      edits[i].mean,
                      math.sqrt(edits[i].varia),
                      ifgs[0], ifgs[i])
+            report[ifgs[0] + ',' + ifgs[i]] = {
+                'length': edits[i].length,
+                'distance-mean': edits[i].mean,
+                'distance-varia': edits[i].varia
+            }
         if confusion:
             for i in range(1, len(ifgs)):
-                LOG.info("most frequent confusion / %s vs %s: %s",
-                         ifgs[0], ifgs[i], self.aligners[i].get_confusion(confusion))
-            
-    def zip_input_files(self, ifgs):
-        """Get a list (for each physical page) of tuples (for each input file group) of METS files."""
-        LOG = getLogger('processor.EvaluateLines')
-        ifts = list() # file tuples
-        pages = dict() # page->file lists
-        for i, ifg in enumerate(ifgs):
-            for file_ in self.workspace.mets.find_all_files(
-                    pageId=self.page_id, fileGrp=ifg, mimetype=MIMETYPE_PAGE):
-                if not file_.pageId:
+                if not edits[i].length:
                     continue
-                LOG.debug("adding page %s to input file group %s", file_.pageId, ifg)
-                ift = pages.setdefault(file_.pageId, [None]*len(ifgs))
-                ift[i] = file_
-        for page, ifiles in pages.items():
-            for i, ifg in enumerate(ifgs):
-                if not ifiles[i]:
-                    # other fallback options?
-                    LOG.error('found no page %s in file group %s',
-                              page, ifg)
-            if ifiles[0]:
-                ifts.append(tuple(ifiles))
-        return ifts
+                conf = self.aligners[i].get_confusion(confusion)
+                LOG.info("most frequent confusion / %s vs %s: %s",
+                         ifgs[0], ifgs[i], conf)
+                report[ifgs[0] + ',' + ifgs[i]]['confusion'] = repr(conf)
+        # write back result to overall report
+        file_id = self.output_file_grp
+        file_path = os.path.join(self.output_file_grp, file_id + '.json')
+        self.workspace.add_file(
+            ID=file_id,
+            file_grp=self.output_file_grp,
+            pageId=None,
+            local_filename=file_path,
+            mimetype='application/json',
+            content=json.dumps(report, indent=2))
 
-def _page_get_lines(pcgts):
+def page_get_lines(pcgts):
     '''Get all TextLines in the page.
     
     Iterate the element hierarchy of the page `pcgts` down
